import random
from spacy.util import minibatch
from .util import cyclic_triangular_rate


def train_while_improving(
    nlp,
    train_data,
    evaluate,
    *,
    learning_rate: float,
    batch_size: int,
    weight_decay: float,
    classifier_lr: float,
    dropout: float,
    lr_range: int,
    lr_period: int,
    steps_per_batch: int,
    patience: int,
    eval_every: int
):
    """Train until an evaluation stops improving. Works as a generator,
    with each iteration yielding a tuple `(batch, info, is_best_checkpoint)`,
    where info is a dict, and is_best_checkpoint is in [True, False, None] --
    None indicating that the iteration was not evaluated as a checkpoint.
    The evaluation is conducted by calling the evaluate callback, which should

    Positional arguments:
        nlp: The spaCy pipeline to evaluate.
        train_data (Sized[Tuple[Union[unicode, Doc], Union[GoldParse, dict]]]):
            The training data.
        evaluate (Callable[[], Tuple[float, Any]]): A callback to perform evaluation.
            The callback should take no arguments and return a tuple
            `(main_score, other_scores)`. The main_score should be a float where
            higher is better. other_scores can be any object.

    The following hyper-parameters (passed as keyword arguments) may need to be
    adjusted for your problem:

        learning_rate (float): Central learning rate to cycle around. 1e-5 is
            often good, but it can depend on the batch size.
        batch_size (int): The number of examples per iteration. Try 32 and 128.
            Larger batch size makes the gradient estimation more accurate, but
            means fewer updates to the model are made per pass over the data.
            With more passes over the data, the updates are less noisy, which
            can lead to inferior generalisation and longer training times.
            The batch size is expressed in number of examples, so the best
            value will depend on the number of words per example in your data.
            If your documents are long, you can use a lower batch size (because
            you're using more information to estimate the gradients). With a
            large dataset and short examples, try a batch size of 128, possibly
            setting a higher value for `steps_per_batch` if you run out of memory
            (see below). Also try a smaller batch size like 32.

    The following hyper-parameters can affect accuracy, but generalize fairly
    well and probably don't need to be tuned:

        weight_decay (float): The weight decay for the AdamW optimizer. 0.005
            is a good value.
        classifier_lr (float): The learning rate for the classifier parameters,
            which must be trained from scatch on each new problem. A value of
            0.001 is good -- it's best for the classifier to learn much faster
            than the rest of the network, which is initialised from the language
            model.
        lr_range (int): The range to vary the learning rate over during training.
            The learning rate will cycle between learning_rate / lr_range and
            learning_rate * lr_range. 2 is good.
        lr_period (int): How many epochs per min-to-max period in the cycle.
            2 is good. Definitely don't set patience < lr_period * 2 --- you
            want at least one full cycle before you give up.

    The following hyper-parameters impact compute budgets.

        patience (int): How many evaluations to allow without improvement
            before giving up. e.g. if patience is 5 and the best evaluation
            was the tenth one, the loop may halt from evaluation 15 onward.
            The loop only halts after a full epoch though, so depending on the
            evaluation frequency, more than `patience` checkpoints may occur
            after the best one. 10 is good.
        eval_every (int): How often to evaluate, in number of iterations.
            max(100, (len(train_data) // batch_size) // 10) is good -- i.e.
            either 100, or about every 10% of a full epoch. For small training
        steps_per_batch (int): Accumulate gradients over a number of steps for
            each batch. This allows you to use a higher batch size with less memory,
            at the expense of potentially slower compute costs. If you don't need
            it, just set it to 1.

    Every iteration, the function yields out a tuple with:

    * batch: A zipped sequence of Tuple[Doc, GoldParse] pairs.
    * info: A dict with various information about the last update (see below).
    * is_best_checkpoint: A value in None, False, True, indicating whether this
        was the best evaluation so far. You should use this to save the model
        checkpoints during training. If None, evaluation was not conducted on
        that iteration. False means evaluation was conducted, but a previous
        evaluation was better.

    The info dict provides the following information:

        epoch (int): How many passes over the data have been completed.
        step (int): How many steps have been completed.
        score (float): The main score form the last evaluation.
        other_scores: : The other scores from the last evaluation.
        loss: The accumulated losses throughout training.
        checkpoints: A list of previous results, where each result is a (score, step, epoch) tuple.
    """
    nr_eg = len(train_data)
    nr_batch = nr_eg // batch_size
    steps_per_epoch = nr_batch * steps_per_batch
    optimizer = nlp.resume_training()
    learn_rates = cyclic_triangular_rate(
        learning_rate / lr_range, learning_rate * lr_range, steps_per_epoch
    )
    optimizer.trf_lr = next(learn_rates)
    optimizer.trf_weight_decay = HP.weight_decay
    # This sets the learning rate for the Thinc layers, i.e. just the final
    # softmax. By keeping this LR high, we avoid a problem where the model
    # spends too long flat, which harms the transfer learning.
    optimizer.alpha = HP.classifier_lr
    epoch = 0
    step = 0
    results = []
    while True:
        random.shuffle(train_data)
        for batch in minibatch(train_data, size=(batch_size // steps_per_batch)):
            optimizer.trf_lr = next(learn_rates)
            docs, golds = zip(*batch)
            losses = {}
            nlp.update(
                docs,
                golds,
                drop=HP.dropout,
                losses=losses,
                sgd=(optimizer if (step % steps_per_batch == 0) else None),
            )
            if step != 0 and not (step % (eval_every * steps_per_batch)):
                with nlp.use_params(optimizer.averages):
                    score, other_scores = evaluate()
                results.append((score, step, epoch))
                is_best_checkpoint = score == max(results)[0]
            else:
                score, other_scores = (None, None)
                is_best_checkpoint = None
            info = {
                "epoch": epoch,
                "step": step,
                "score": score,
                "other_scores": other_scores,
                "loss": losses,
                "checkpoints": results,
            }
            yield batch, info, is_best_checkpoint
            step += 1
        epoch += 1
        # Stop if no improvement in HP.patience checkpoints
        if results:
            best_score, best_step, best_epoch = max(results)
            if ((step - best_step) // HP.eval_every) >= HP.patience:
                break



from spacy.language import Language
from spacy.tokens import Doc, Span, Token
from spacy.util import get_lang_class
from spacy.gold import GoldParse
from .util import is_special_token, pkg_meta, ATTRS, PIPES, LANG_FACTORY


class TransformersLanguage(Language):
    """A subclass of spacy.Language that holds a Transformer pipeline.

    Transformer pipelines work only slightly differently from spaCy's default
    pipelines. Specifically, we introduce a new pipeline component at the start
    of the pipeline, TransformerTok2Vec. We then modify the nlp.update()
    function to run the TransformerTok2Vec before the other pipeline components,
    and backprop it after the other components are done.
    """

    lang_factory_name = LANG_FACTORY

    @staticmethod
    def install_extensions():
        tok2vec_attrs = [
            ATTRS.last_hidden_state,
            ATTRS.pooler_output,
            ATTRS.all_hidden_states,
            ATTRS.all_attentions,
            ATTRS.d_last_hidden_state,
            ATTRS.d_pooler_output,
            ATTRS.d_all_hidden_states,
            ATTRS.d_all_attentions,
        ]
        for attr in tok2vec_attrs:
            Doc.set_extension(attr, default=None)
            Span.set_extension(attr, getter=get_span_tok2vec_getter(attr))
            Token.set_extension(attr, getter=get_token_tok2vec_getter(attr))
        wp_attrs = [ATTRS.alignment, ATTRS.word_pieces, ATTRS.word_pieces_]
        for attr in wp_attrs:
            Doc.set_extension(attr, default=None)
            Span.set_extension(attr, getter=get_span_wp_getter(attr))
            Token.set_extension(attr, getter=get_token_wp_getter(attr))
        Doc.set_extension(ATTRS.separator, default=None)
        Span.set_extension(
            ATTRS.separator, getter=lambda span: span.doc._.get(ATTRS.separator)
        )
        Token.set_extension(
            ATTRS.separator, getter=lambda token: token.doc._.get(ATTRS.separator)
        )
        Doc.set_extension(ATTRS.segments, getter=get_segments)
        Span.set_extension(ATTRS.segments, getter=get_segments)
        for cls in [Token, Span, Doc]:
            cls.set_extension(ATTRS.start, getter=get_wp_start)
            cls.set_extension(ATTRS.end, getter=get_wp_end)

    def __init__(
        self, vocab=True, make_doc=True, max_length=10 ** 6, meta={}, **kwargs
    ):
        """Initialize the language class. Expects either a trf_name setting in
        the meta or as a keyword argument, specifying the pre-trained model
        name. This is used to set up the model-specific tokenizer.
        """
        meta = dict(meta)
        meta["lang_factory"] = self.lang_factory_name
        # Add this package to requirements to it will be included in the
        # install_requires of any model using this language class
        package = f"{pkg_meta['title']}>={pkg_meta['version']}"
        meta.setdefault("requirements", []).append(package)
        self.lang = meta.get("lang", "xx")
        self.Defaults = get_defaults(self.lang)
        super().__init__(vocab, make_doc, max_length, meta=meta, **kwargs)

    def update(self, docs, golds, drop=0.0, sgd=None, losses=None, component_cfg={}):
        component_cfg = dict(component_cfg)
        if self.has_pipe("sentencizer"):
            sentencizer = self.get_pipe("sentencizer")
        else:
            sentencizer = lambda doc: doc
        if self.has_pipe(PIPES.wordpiecer):
            wp = self.get_pipe(PIPES.wordpiecer)
        else:
            wp = lambda doc: doc
        tok2vec = self.get_pipe(PIPES.tok2vec)
        new_docs = []
        new_golds = []
        for doc, gold in zip(docs, golds):
            if isinstance(doc, str):
                doc = self.make_doc(doc)
            doc = sentencizer(doc)
            if doc._.get(ATTRS.word_pieces) is None:
                doc = wp(doc)
            if not isinstance(gold, GoldParse):
                gold = GoldParse(doc, **gold)
            new_docs.append(doc)
            new_golds.append(gold)
        docs = new_docs
        golds = new_golds
        outputs, backprop_tok2vec = tok2vec.begin_update(
            docs, drop=drop, **component_cfg.get(PIPES.tok2vec, {})
        )
        tok2vec.set_annotations(docs, outputs)
        for doc in docs:
            assert doc._.get(ATTRS.last_hidden_state) is not None
        with self.disable_pipes(PIPES.tok2vec):
            super().update(
                docs,
                golds,
                drop=0.1,
                sgd=sgd,
                losses=losses,
                component_cfg=component_cfg,
            )
        backprop_tok2vec(docs, sgd=sgd)

    def resume_training(self, sgd=None, component_cfg=None, **kwargs):
        """Continue training a pre-trained model.

        Before running the normal Language.resume_training method, we do the
        following:

        * Look for a tok2vec pipeline component. The component name can be
            changed with the tok2vec_name keyword
            argument. If no component is found, a ValueError is raised.
        * If any other components have `component.model == True` and a
            `.begin_training()` method, we call the `.begin_training()` method.
            Configuration can be passed in using the component_cfg keyword
            argument. If unset, we also pass in a value for token_vector_width,
            which we read from the tok2vec component.
        """
        if component_cfg is None:
            component_cfg = {}
        tok2vec_name = kwargs.get("tok2vec_name", PIPES.tok2vec)
        tok2vec = self.get_pipe(tok2vec_name)
        token_vector_width = tok2vec.token_vector_width
        for name, component in self.pipeline:
            if name == tok2vec_name:
                continue
            elif getattr(component, "model", None) is not True:
                continue
            elif not hasattr(component, "begin_training"):
                continue
            cfg = component_cfg.get(name, {})
            if "tok2vec_name" not in component_cfg:
                cfg["tok2vec_name"] = tok2vec_name
            if "token_vector_width" not in component_cfg:
                cfg["token_vector_width"] = token_vector_width
            component.cfg.update(cfg)
            component.begin_training(pipeline=self.pipeline, sgd=False, **cfg)
            assert component.model is not True
        optimizer = super().resume_training(sgd=sgd, **kwargs)
        optimizer.L2 = 0.0
        return optimizer


def get_defaults(lang):
    """Get the language-specific defaults, if available in spaCy."""
    try:
        lang_cls = get_lang_class(lang)
        return lang_cls.Defaults
    except ImportError:
        return Language.Defaults


def get_wp_start(span):
    if isinstance(span, Token):
        span = span.doc[span.i : span.i + 1]
    for token in span:
        if token._.get(ATTRS.alignment):
            wp_start = token._.get(ATTRS.alignment)[0]
            break
    else:
        return None
    wordpieces = span.doc._.get(ATTRS.word_pieces_)
    # This is a messy way to check for the XLNet-style pattern, where we can
    # have <sep> <cls>. In the BERT-style pattern, we have [cls] at start.
    if is_special_token(wordpieces[0]):
        if wp_start >= 1 and is_special_token(wordpieces[wp_start - 1]):
            return wp_start - 1
    return wp_start


def get_wp_end(span):
    if isinstance(span, Token):
        span = span.doc[span.i : span.i + 1]
    for token in reversed(span):
        if token._.get(ATTRS.alignment):
            wp_end = token._.get(ATTRS.alignment)[-1]
            break
    else:
        return None
    wordpieces = span.doc._.get(ATTRS.word_pieces_)
    if (wp_end + 1) < len(wordpieces) and is_special_token(wordpieces[wp_end + 1]):
        wp_end += 1
    # This is a messy way to check for the XLNet-style pattern, where we can
    # have <sep> <cls>. In the BERT-style pattern, we have [cls] at start.
    if not is_special_token(wordpieces[0]):
        if (wp_end + 1) < len(wordpieces) and is_special_token(wordpieces[wp_end + 1]):
            wp_end += 1
    return wp_end


def get_span_wp_getter(attr):
    def span_alignment_getter(span):
        return [token._.get(attr) for token in span]

    def span_getter(span):
        start = span._.get(ATTRS.start)
        end = span._.get(ATTRS.end)
        if start is None and end is None:
            return []
        doc_values = span.doc._.get(attr)
        start = start if start is not None else 0
        if end is None:
            return doc_values[start:]
        return doc_values[start : end + 1]

    if attr == ATTRS.alignment:
        return span_alignment_getter
    else:
        return span_getter


def get_token_wp_getter(attr):
    def token_alignment_getter(token):
        doc_values = token.doc._.get(attr)
        return doc_values[token.i] if doc_values is not None else None

    def token_wordpiece_getter(token):
        doc_values = token.doc._.get(attr)
        start = token._.get(ATTRS.start)
        end = token._.get(ATTRS.end)
        if start is None and end is None:
            return []
        return [doc_values[i] for i in range(start, end + 1)]

    if attr == ATTRS.alignment:
        return token_alignment_getter
    else:
        return token_wordpiece_getter


def get_span_tok2vec_getter(attr):
    def span_getter(span):
        doc_activations = span.doc._.get(attr)
        if doc_activations is None:
            return None
        wp_start = span[0]._.get(ATTRS.start)
        wp_end = span[-1]._.get(ATTRS.end)
        if wp_start is not None and wp_end is not None:
            return doc_activations[wp_start : wp_end + 1]
        else:
            # Return empty slice.
            return doc_activations[0:0]

    return span_getter


def get_token_tok2vec_getter(attr):
    def token_getter(token):
        # Delegate through span, so get a span with just the token.
        span = token.doc[token.i : token.i + 1]
        return span._.get(attr)

    return token_getter


def get_segments(doc):
    separator = doc._.get(ATTRS.separator)
    if separator is not None:
        start = 0
        for token in doc:
            if token.text == separator:
                yield doc[start : token.i + 1]
                start = token.i + 1
        yield doc[start:]
    else:
        yield doc[:]



import spacy.pipeline
from ..model_registry import get_model_function
from ..util import PIPES

DEBUG_LOSS = False


class TransformersEntityRecognizer(spacy.pipeline.EntityRecognizer):
    """Subclass of spaCy's built-in EntityRecognizer component that supports
    using the features assigned by the PyTorch-Transformers models via the token
    vector encoder. It requires the TransformerTokenVectorEncoder to run before
    it in the pipeline.
    """

    name = PIPES.ner
    factory = PIPES.ner

    @classmethod
    def from_nlp(cls, nlp, **cfg):
        """Factory to add to Language.factories via entry point."""
        return cls(nlp.vocab, **cfg)

    @classmethod
    def Model(cls, nr_class, **cfg):
        hidden_width = cfg.get("hidden_width", 64)
        token_vector_width = cfg.get("token_vector_width", 96)
        nr_feature = cls.nr_feature
        maxout_pieces = cfg.get("maxout_pieces", 3)
        tensor_size = cfg.get("tensor_size", 768)

        configs = {
            "tok2vec": {
                "arch": "tensor_affine_tok2vec",
                "config": {
                    "output_size": token_vector_width,
                    "tensor_size": tensor_size
                }
            },
            "lower": {
                "arch": "precomputable_maxout",
                "config": {
                    "output_size": hidden_width,
                    "input_size": token_vector_width,
                    "number_features": nr_feature,
                    "maxout_pieces": maxout_pieces
                }
            },
            "upper": {
                "arch": "affine_output",
                "config": {
                    "output_size": nr_class,
                    "input_size": hidden_width,
                    "drop_factor": 0.0
                }
            }`
        }
        tok2vec_arch = get_model_function(configs["tok2vec"]["arch"])
        lower_arch = get_model_function(configs["lower"]["arch"])
        upper_arch = get_model_function(configs["upper"]["arch"])
        tok2vec = tok2vec_arch(**configs["tok2vec"]["config"])
        lower = lower_arch(**configs["lower"]["config"])

        with Model.use_device('cpu'):
            upper = upper_arch(**configs["upper"]["config"])
        upper.W *= 0
        return ParserModel(tok2vec, lower, upper), configs



import pytest
from spacy_transformers import TransformersWordPiecer
from spacy_transformers.util import ATTRS
from spacy.tokens import Doc
from spacy.vocab import Vocab


@pytest.fixture
def vocab():
    return Vocab()


@pytest.fixture
def wordpiecer(name):
    return TransformersWordPiecer.from_pretrained(vocab, trf_name=name)


def test_alignment_extension_attr(vocab):
    doc = Doc(vocab, words=["hello", "world", "test"])
    doc._.set(ATTRS.alignment, [[1, 2], [3, 4], [5, 6]])
    assert doc[0]._.get(ATTRS.alignment) == [1, 2]
    assert doc[1]._.get(ATTRS.alignment) == [3, 4]
    assert doc[2]._.get(ATTRS.alignment) == [5, 6]
    assert doc[0:2]._.get(ATTRS.alignment) == [[1, 2], [3, 4]]
    assert doc[1:3]._.get(ATTRS.alignment) == [[3, 4], [5, 6]]


def test_wp_span_extension_attr(name, vocab, wordpiecer):
    if name == "gpt2":
        return
    doc = Doc(vocab, words=["hello", "world"])
    for w in doc[1:]:
        w.is_sent_start = False
    doc = wordpiecer(doc)
    assert len(doc._.get(ATTRS.word_pieces)) == 4
    assert doc[0]._.get(ATTRS.start) == 0
    assert doc[-1]._.get(ATTRS.end) == 3
    assert len(doc[:]._.get(ATTRS.word_pieces)) == 4



import pytest
from spacy_transformers import TransformersWordPiecer
from spacy_transformers.util import is_special_token, get_tokenizer, ATTRS
from spacy.vocab import Vocab
from spacy.tokens import Doc
from spacy.pipeline import Sentencizer


@pytest.fixture(scope="session")
def wp(name):
    return TransformersWordPiecer.from_pretrained(Vocab(), trf_name=name)


@pytest.fixture(scope="session")
def sentencizer():
    return Sentencizer()


def test_wordpiecer(wp):
    words = ["hello", "world", "this", "is", "a", "test"]
    doc = Doc(wp.vocab, words=words)
    doc[0].is_sent_start = True
    doc[1].is_sent_start = False
    doc = wp(doc)
    cleaned_words = [wp.model.clean_wp_token(t) for t in doc._.get(ATTRS.word_pieces_)]
    cleaned_words = [w for w in cleaned_words if not is_special_token(w)]
    assert "".join(cleaned_words) == "".join(words)


@pytest.mark.parametrize(
    "words,target_name,expected_align",
    [
        (
            ["hello", "world", "this", "is", "a", "teest"],
            "bert-base-uncased",
            [[1], [2], [3], [4], [5], [6, 7]],
        ),
        (
            ["hello", "world", "this", "is", "a", "teest"],
            "xlnet-base-cased",
            [[0], [1], [2], [3], [4], [5, 6]],
        ),
        (["å\taa", ".", "が\nπ"], "bert-base-uncased", [[1, 2], [3], [6, 7]]),
        # no idea why this fails
        pytest.param(["å\taa", ".", "が\nπ"], "xlnet-base-cased", [[0, 1, 2], [4], [7, 8, 10]], marks=pytest.mark.xfail()),
        (["\u3099"], "bert-base-uncased", [[]]),
        (["I.\n\n\n\n\n"], "bert-base-uncased", [[1, 2]]),
        # max length of 512 minus 2 special tokens -> 510 aligned tokens,
        # remaining truncated per sentence
        (["x"] * 599 + ["."] + ["y"] * 600, "bert-base-uncased", [[x] for x in range(1, 511)] + [[]] * 90 + [[x] for x in range(513, 1023)] + [[]] * 90),
    ],
)
def test_align(wp, sentencizer, name, words, target_name, expected_align):
    if name != target_name:
        pytest.skip()
    doc = Doc(wp.vocab, words=words)
    doc = sentencizer(doc)
    doc = wp(doc)
    assert doc._.get(ATTRS.alignment) == expected_align


def test_xlnet_weird_align(name, wp):
    if "xlnet" not in name.lower():
        pytest.skip()
    text = "Well, i rented this movie and found out it realllllllly sucks."
    spacy_tokens = [
        "Well",
        ",",
        "i",
        "rented",
        "this",
        "movie",
        "and",
        "found",
        "out",
        "it",
        "realllllllly",
        "sucks",
        ".",
    ]
    spaces = [True] * len(spacy_tokens)
    spaces[0] = False
    spaces[-2] = False
    spaces[-1] = False
    doc = Doc(wp.vocab, words=spacy_tokens, spaces=spaces)
    doc[1].is_sent_start = False
    assert doc.text == text
    doc = wp(doc)
    assert doc._.get(ATTRS.word_pieces_)[-2] == "</s>"
    assert doc._.get(ATTRS.word_pieces_)[-1] == "<cls>"


def test_tokenizers_to_from_bytes(name):
    text = "hello world"
    tokenizer_cls = get_tokenizer(name)
    tokenizer = tokenizer_cls.from_pretrained(name)
    doc = tokenizer.tokenize(text)
    assert isinstance(doc, list) and len(doc)
    bytes_data = tokenizer.to_bytes(name)
    new_tokenizer = tokenizer_cls.blank().from_bytes(bytes_data)
    new_doc = new_tokenizer.tokenize(text)
    assert isinstance(new_doc, list) and len(new_doc)
    assert doc == new_doc



from .language import TransformersLanguage
from .pipeline.tok2vec import TransformersTok2Vec  # noqa
from .pipeline.textcat import TransformersTextCategorizer  # noqa
from .pipeline.wordpiecer import TransformersWordPiecer  # noqa
from .model_registry import register_model, get_model_function  # noqa
from .util import pkg_meta

__version__ = pkg_meta["version"]
TransformersLanguage.install_extensions()



import pytest
from spacy_transformers.pipeline import TransformersTextCategorizer
from spacy_transformers.util import PIPES
from spacy.gold import GoldParse


def textcat(name, nlp, request):
    arch = request.param
    tensor_size = nlp.get_pipe(PIPES.tok2vec).model.nO
    ner = TransformersEntityRecognizer(nlp.vocab)
    ner.add_label("PERSON")
    config = {"trf_name": name, "tensor_size": tensor_size}
    ner.begin_training(**config)
    return ner

@pytest.mark.xfail
def test_ner_init(nlp):
    ner = TransformersEntityRecognizer(nlp.vocab)
    assert textcat.labels == ()
    textcat.add_label("PERSON")
    assert textcat.labels == ("PERSON",)


@pytest.mark.xfail
def test_ner_call(ner, nlp):
    doc = nlp("hello world")


@pytest.mark.xfail
def test_ner_update(ner, nlp):
    doc = nlp("hello matt")
    optimizer = nlp.resume_training()
    ents = ["O", "U-PERSON"]
    losses = {}
    ner.update([doc], [GoldParse(doc, entities=ents)], sgd=optimizer, losses=losses)
    assert PIPES.ner in losses


@pytest.mark.xfail
def test_ner_update_multi_sentence(ner, nlp):
    doc = nlp("Hello matt. This is ian.")
    assert len(list(doc.sents)) == 2
    optimizer = nlp.resume_training()
    ents = ["O", "U-PERSON", "O", "O", "O", "U-PERSON", "O"]
    losses = {}
    ner.update([doc], [GoldParse(doc, entities=ents)], sgd=optimizer, losses=losses)
    assert PIPES.ner in losses


@pytest.mark.xfail
def test_ner_update_batch(ner, nlp):
    doc1 = nlp("Hello world. This is sentence 2.")
    doc2 = nlp("Hi again. This is sentence 4.")
    ents1 = ["O"] * len(doc1)
    ents2 = ["O"] * len(doc2)
    assert len(list(doc1.sents)) == 2
    assert len(list(doc2.sents)) == 2
    optimizer = nlp.resume_training()
    golds = [GoldParse(doc1, entities=ents1), GoldParse(doc2, entities=ents2)]
    losses = {}
    ner.update([doc1, doc2], golds, sgd=optimizer, losses=losses)
    assert PIPES.ner in losses
